package net.lemnik.eodsql;

import java.util.List;
import java.util.Map;

/**
 * This interface can be used as a mix-in for SQL queries and statements created
 * dynamically at runtime. Usage is like this:
 * 
 * <pre>
 * <code>
 * public interface MyQuery extends BaseQuery, DynamicQuery {
 *     @Select(sql = "SELECT * FROM persons")
 *     public DataSet&lt;Address&gt; list();
 * 
 *     @Select(sql = "SELECT * FROM persons")
 *     public DataSet&lt;Map&lt;String,Object&gt;&gt; listMap();
 * }
 * ...
 * MyQuery query = QueryTool.getQuery(conn, MyQuery.class);
 * DataSet&lt;Map&lt;String,Object&gt;&gt; data = query.select("SELECT * FROM persons WHERE last_name = 'Smith'");
 * </code>
 * </pre>
 * <p>
 * <code>DynamicQuery</code> can be mixed-in to {@link BaseQuery} and
 * {@link TransactionQuery}. The interfaces {@link DynamicBaseQuery} and
 * {@link DynamicTransactionQuery} can be used for that.
 * 
 * @author Bernd Rinn
 */
public interface DynamicQuery {

    /**
     * Performs a SQL query. The returned <code>DataSet</code> is connected and
     * updateable.
     * 
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>DataSet</code>; each row is represented
     *         as one <code>Map&lt;String,Object&gt;</code>.
     */
    DataSet<Map<String, Object>> select(final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Performs a SQL query. The returned <code>DataIterator</code> is connected
     * and read-only.
     * 
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>DataIterator</code>; each row is
     *         represented as one <code>Map&lt;String,Object&gt;</code>.
     */
    DataIterator<Map<String, Object>> selectIterator(final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Performs a SQL query. The returned <code>List</code> is
     * independent of the database result set.
     * 
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>List</code>; each row is
     *         represented as one <code>Map&lt;String,Object&gt;</code>.
     */
    List<Map<String, Object>> selectList(final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Performs a SQL query that returns zero or one row. It is an error
     * condition if the query returns more than one row in its result set.
     * 
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result row as a <code>Map&lt;String,Object&gt;</code>, or
     *         <code>null</code>, if no row was found.
     */
    Map<String, Object> selectRow(final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Performs a SQL query. The returned <code>DataIterator</code> is connected
     * and read-only.
     * 
     * @param fetchSize The fetch size hint to give to the statement.
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>DataIterator</code>; each row is
     *         represented as one <code>Map&lt;String,Object&gt;</code>.
     */
    DataIterator<Map<String, Object>> selectIterator(final int fetchSize,
            final String query, final Object... parameters)
            throws InvalidDataTypeException, InvalidQueryException,
            EoDException;

    /**
     * Performs a SQL query. The returned <code>List</code> is
     * independent of the database result set.
     * 
     * @param fetchSize The fetch size hint to give to the statement.
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>List</code>; each row is
     *         represented as one <code>Map&lt;String,Object&gt;</code>.
     */
    List<Map<String, Object>> selectList(final int fetchSize, final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Performs a SQL query. The returned <code>DataSet</code> is connected and
     * updateable.
     * 
     * @param type The Java type to return one rows in the returned
     *        <code>DataSet</code>.
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>DataSet</code>; each row is represented
     *         as one <code>Map&lt;String,Object&gt;</code>.
     */
    <T> DataSet<T> select(final Class<T> type, final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Performs a SQL query. The returned <code>DataIterator</code> is connected
     * and read-only.
     * 
     * @param type The Java type to return one rows in the returned
     *        <code>DataIterator</code>.
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>DataIterator</code>; each row is
     *         represented as an object of type <code>T</code>.
     */
    <T> DataIterator<T> selectIterator(final Class<T> type, final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Performs a SQL query. The returned <code>List</code> is
     * independent of the database result set.
     * 
     * @param type The Java type to return one rows in the returned
     *        <code>List</code>.
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>List</code>; each row is
     *         represented as an object of type <code>T</code>.
     */
    <T> List<T> selectList(final Class<T> type, final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Performs a SQL query that returns zero or one row. It is an error
     * condition if the query returns more than one row in its result set.
     * 
     * @param type The Java type to return one rows in the returned
     *        <code>DataSet</code>.
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result row as an object of type <code>T</code>, or
     *         <code>null</code>, if no row was found.
     */
    <T> T selectRow(final Class<T> type, final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Performs a SQL query.
     * 
     * @param params The parameters of the query.
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>DataSet</code>; each row is represented
     *         as one <code>Map&lt;String,Object&gt;</code>.
     */
    <T> DataSet<T> select(final QueryParams<T> params, final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Performs a SQL query. The returned <code>DataIterator</code> is connected
     * and read-only.
     * 
     * @param type The Java type to return one rows in the returned
     *        <code>DataIterator</code>.
     * @param fetchSize The fetch size hint to give to the statement.
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>DataIterator</code>; each row is
     *         represented as an object of type <code>T</code>.
     */
    <T> DataIterator<T> selectIterator(final Class<T> type,
            final int fetchSize, final String query, final Object... parameters)
            throws InvalidDataTypeException, InvalidQueryException,
            EoDException;

    /**
     * Performs a SQL query. The returned <code>List</code> is
     * independent of the database result set.
     * 
     * @param type The Java type to return one rows in the returned
     *        <code>List</code>.
     * @param fetchSize The fetch size hint to give to the statement.
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The result set as <code>List</code>; each row is
     *         represented as an object of type <code>T</code>.
     */
    <T> List<T> selectList(final Class<T> type, final int fetchSize, final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Executes a SQL statement.
     * 
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The number of rows updated by the SQL statement, or -1 if not
     *         applicable. <b>Note:</b> Not all JDBC drivers support this
     *         cleanly.
     */
    int update(final String query, final Object... parameters)
            throws InvalidDataTypeException, InvalidQueryException,
            EoDException;

    /**
     * Executes a SQL statement as a batch for all parameter values provided.
     * 
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template. At
     *        least one of the parameters needs to be an array or
     *        <code>Collection</code>. If multiple parameters are arrays or
     *        <code>Collection</code>, all of them need to have the same size.
     * 
     * @return The number of rows updated by the SQL statement, or -1 if not
     *         applicable. <b>Note:</b> Not all JDBC drivers support this
     *         cleanly.
     */
    int batchUpdate(final String query, final Object... parameters)
            throws InvalidDataTypeException, InvalidQueryException,
            EoDException;

    /**
     * Executes a SQL statement. Supposed to be used for INSERT statements with
     * an automatically generated integer key.
     * 
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The automatically generated key. <b>Note:</b> Not all JDBC
     *         drivers support this cleanly.
     */
    long insert(final String query, final Object... parameters)
            throws InvalidDataTypeException, InvalidQueryException,
            EoDException;

    /**
     * Executes a SQL statement. Supposed to be used for INSERT statements with
     * one or more automatically generated keys.
     * 
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template.
     * 
     * @return The automatically generated keys. <b>Note:</b> Not all JDBC
     *         drivers support this cleanly and it is very driver-dependent what
     *         keys are present in the returned map.
     */
    Map<String, Object> insertMultiKeys(final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;

    /**
     * Executes a SQL statement as a batch for all parameter values provided.
     * Supposed to be used for INSERT statements with an automatically generated
     * integer key.
     * 
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template. At
     *        least one of the parameters needs to be an array or
     *        <code>Collection</code>. If multiple parameters are arrays or
     *        <code>Collection</code>, all of them need to have the same size.
     * 
     * @return The automatically generated key for each element of the batch.
     *         <b>Note:</b> Not all JDBC drivers support this cleanly.
     */
    long[] batchInsert(final String query, final Object... parameters)
            throws InvalidDataTypeException, InvalidQueryException,
            EoDException;

    /**
     * Executes a SQL statement as a batch for all parameter values provided.
     * Supposed to be used for INSERT statements with one or more automatically
     * generated keys.
     * 
     * @param query The SQL query template.
     * @param parameters The parameters to fill into the SQL query template. At
     *        least one of the parameters needs to be an array or
     *        <code>Collection</code>. If multiple parameters are arrays or
     *        <code>Collection</code>, all of them need to have the same size.
     * 
     * @return The automatically generated keys for each element of the batch.
     *         <b>Note:</b> Not all JDBC drivers support this cleanly and it is
     *         very driver-dependent what keys are present in the returned map.
     */
    Map<String, Object>[] batchInsertMultiKeys(final String query,
            final Object... parameters) throws InvalidDataTypeException,
            InvalidQueryException, EoDException;
}
