package net.lemnik.eodsql;

/**
 * Class to encode the query parameters used in a {@link DynamicQuery}.
 * 
 * @author Bernd Rinn
 */
public class QueryParams<T> {

    protected Class<T> baseReturnType;

    private boolean readOnly;

    private boolean disconnected;

    private int fetchSize;

    protected QueryParams() {
    }

    /**
     * Creates a new {@QueryParams} object.
     *
     * @param baseReturnType The return type that represents a single row in the query result.
     */
    public QueryParams(Class<T> baseReturnType) {
        this.baseReturnType = baseReturnType;
    }

    /**
     * Sets whether the returned {@link DataSet} will be read-only (default is
     * <code>false</code>).
     * <p>
     * This has no effect on queries that return a {@link DataIterator} or
     * {@link java.util.List} or a single row as they are always read-only.
     * 
     * @return This object, for chaining calls.
     */
    public QueryParams<T> readOnly(boolean readOnly) {
        this.readOnly = readOnly;
        return this;
    }

    /**
     * Sets whether the returned {@link DataSet} will be disconnected (default
     * is <code>false</code>).
     * <p>
     * This has no effect on queries that return a {@link DataIterator} or
     * {@link java.util.List} or a single row. {@link DataIterator}s are always
     * connected, {@link java.util.List} and single row results are always
     * disconnected.
     * 
     * @return This object, for chaining calls.
     */
    public QueryParams<T> disconnected(boolean disconnected) {
        this.disconnected = disconnected;
        if (disconnected) {
            // Disconnected datasets may not be updateable.
            this.readOnly = true;
        }
        return this;
    }

    /**
     * Sets a fetch size for the query. The default is 0 which leaves the decision to the JDBC driver.
     * 
     * @return This object, for chaining calls.
     */
    public QueryParams<T> fetchSize(int fetchSize) {
        this.fetchSize = fetchSize;
        return this;
    }

    public Class<T> getBaseReturnType() {
        return baseReturnType;
    }

    public boolean isReadOnly() {
        return readOnly;
    }

    public boolean isDisconnected() {
        return disconnected;
    }

    public int getFetchSize() {
        return fetchSize;
    }

}
