package net.lemnik.eodsql.impl;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import net.lemnik.eodsql.Call;
import net.lemnik.eodsql.DataIterator;
import net.lemnik.eodsql.DataSet;
import net.lemnik.eodsql.InvalidQueryException;
import net.lemnik.eodsql.QueryTool;
import net.lemnik.eodsql.Select;
import net.lemnik.eodsql.TypeMapper;
import net.lemnik.eodsql.Update;
import net.lemnik.eodsql.spi.Context;
import net.lemnik.eodsql.spi.MethodImplementation;
import net.lemnik.eodsql.spi.StatementResource;
import net.lemnik.eodsql.spi.util.Query;
import net.lemnik.eodsql.spi.util.ResultSetWrapper;

/**
 * Created on 2008/06/28
 * @author Jason Morris
 */
abstract class AbstractMethodImplementation<A extends Annotation>
        implements MethodImplementation<A> {

    private final static List<Integer> ALL_RESULT_SET_TYPES = Arrays.asList(
            ResultSet.TYPE_SCROLL_SENSITIVE,
            ResultSet.TYPE_SCROLL_INSENSITIVE,
            ResultSet.TYPE_FORWARD_ONLY                                        
        );

    protected ResultSetWrapper<?> wrapper = null;

    protected Query query = null;

    private TypeMapper<?>[] parameterMappers = null;

    protected PreparedStatement createPreparedStatement(
            final Context<A> context)
            throws SQLException {

        final Connection connection = context.getResource(Connection.class).get();
        int preferredResultSetType = wrapper.getPreferredResultSetType();
        int resultSetType = getResultSetType(connection.getMetaData(), preferredResultSetType);
        if (supportsScrolling(preferredResultSetType) && false == supportsScrolling(resultSetType)) {
            if (isDisconnected(context)) {
                resultSetType = ResultSet.TYPE_FORWARD_ONLY;
            } else {
                final String dbName = connection.getMetaData().getDatabaseProductName();
                throw new SQLException("This database (" + dbName 
                                + ") does not support scrollable cursors, thus connected DataSets are not available.");
            }
        }
        final PreparedStatement statement = connection.prepareStatement(
                query.toString(),
                resultSetType,
                wrapper.getPreferredResultSetConcurrency());
        final int queryTimeout = getQueryTimeout(context);
        if (queryTimeout > 0) {
            try {
                statement.setQueryTimeout(queryTimeout);
            } catch (Exception ex) {
                // Some JDBC drivers don't support setting a query timeout - keep quiet about it.
            }
        }

        context.setResource(new StatementResource(statement));

        return statement;
    }

    private int getResultSetType(DatabaseMetaData dbMetaData, int preferredResultSetType) throws SQLException {
        if (dbMetaData.supportsResultSetType(preferredResultSetType)) {
            return preferredResultSetType;
        }
        for (int type : ALL_RESULT_SET_TYPES) {
            if (dbMetaData.supportsResultSetType(type)) {
                return type;
            }
        }
        // We assume that the dbMetaData.supportsResultSetType() call lied to us and that the most 
        // basic result set is supported.
        return ResultSet.TYPE_FORWARD_ONLY;
    }

    private boolean supportsScrolling(int resultSetType) {
        return resultSetType == ResultSet.TYPE_SCROLL_SENSITIVE
                || resultSetType == ResultSet.TYPE_SCROLL_INSENSITIVE;
    }

    private boolean isDisconnected(final Context<A> context) {
        if (context.getAnnotation() instanceof Select
                && ((Select) context.getAnnotation()).disconnected()) {
            return true;
        }
        if (DataSet.class.isAssignableFrom(context.getReturnType())
                || DataIterator.class.isAssignableFrom(context.getReturnType())) {
            return false;
        }
        return true;
    }

    private int getQueryTimeout(final Context<A> context) {
        if (context.getAnnotation() instanceof Select) {
            final Select select = (Select) context.getAnnotation();
            if (select.timeoutParameterIndex() > 0
                    && context.getParameters()[select.timeoutParameterIndex() - 1] != null) {
                return (Integer) context.getParameters()[select.timeoutParameterIndex() - 1];
            }
            return select.timeout();
        }
        if (context.getAnnotation() instanceof Update) {
            final Update update = (Update) context.getAnnotation();
            if (update.timeoutParameterIndex() > 0
                    && context.getParameters()[update.timeoutParameterIndex() - 1] != null) {
                return (Integer) context.getParameters()[update.timeoutParameterIndex() - 1];
            }
            return update.timeout();
        }
        if (context.getAnnotation() instanceof Call) {
            final Call call = (Call) context.getAnnotation();
            if (call.timeoutParameterIndex() > 0
                    && context.getParameters()[call.timeoutParameterIndex() - 1] != null) {
                return (Integer) context.getParameters()[call.timeoutParameterIndex() - 1];
            }
            return call.timeout();
        }
        return 0;
    }

    protected void fillPreparedStatementParameters(final Context<?> context,
            final PreparedStatement statement) throws SQLException {

        final TypeMapper<?>[] mappers = getParameterMappers();

        for (int i = 0; i < mappers.length; i++) {
            final Object parameter = query.getParameter(context, i);

            @SuppressWarnings("unchecked")
            final TypeMapper<Object> parameterMapper =
                    (TypeMapper<Object>)mappers[i];

            parameterMapper.set(statement, i + 1, parameter);
        }
    }

    protected TypeMapper<?>[] getParameterMappers() {
        TypeMapper[] mappers = parameterMappers;

        if (mappers == null) {
            mappers = getParameterTypeMappers(query, null);
            parameterMappers = mappers;
        }

        return mappers;
    }

    protected void setParameterMappers(final Class<? extends TypeMapper>[] customTypeMapperClasses) {
        parameterMappers = getParameterTypeMappers(query, customTypeMapperClasses);
    }

    protected static TypeMapper<?>[] getParameterTypeMappers(final Query query,
            final Class<? extends TypeMapper>[] customTypeMapperClasses) {
        final int parameters = query.getParameterCount();
        final TypeMapper<?>[] mappers = new TypeMapper<?>[parameters];
        final Map<Class, TypeMapper> knownMappers = QueryTool.getTypeMap();

        for (int i = 0; i < parameters; i++) {
            final Constructor<? extends TypeMapper> customMapperConstructor 
                = getConstructor(customTypeMapperClasses, i);
            if (customMapperConstructor != null) {
                try {
                    mappers[i] = customMapperConstructor.newInstance();
                } catch (Exception ex) {
                    throw new InvalidQueryException("Cannot construct custom type mapper " 
                                    + customMapperConstructor.getName(), ex);
                }
            } else {
            final TypeMapper mapper = knownMappers.get(query.getParameterType(i));

                if (mapper != null) {
                    mappers[i] = mapper;
                } else {
                    throw new InvalidQueryException("Unknown primitive type: " +
                        query.getParameterType(i).getName());
                }
            }
        }

        return mappers;
    }

    private static Constructor<? extends TypeMapper> getConstructor(
            Class<? extends TypeMapper>[] customTypeMapperClasses, int i) {
        if (customTypeMapperClasses == null || i >= customTypeMapperClasses.length) {
            return null;
        }
        final Class<? extends TypeMapper> customTypeMapperClass = customTypeMapperClasses[i];
        if (customTypeMapperClass.isInterface()) {
            return null;
        }
        try {
            return customTypeMapperClass.getConstructor();
        } catch (Exception ex) {
            throw new IllegalArgumentException("TypeMapper classes must have a default " +
                    "(null) constructor", ex);
        }
    }

}
