package net.lemnik.eodsql.impl;

import java.sql.Connection;
import java.util.HashMap;
import java.util.Map;

/**
 * A store for managed connections. A "managed connection" is a connection being
 * provided by a transaction manager. No transaction management or closing
 * should be done on them. Database connections are per thread.
 * <p>
 * <b>This is an internal implementation class and should not be used by code
 * outside EoDSQL.</b>
 * 
 * @author Bernd Rinn
 */
public class ManagedConnections {

    /** The default database. */
    public static final String DEFAULT_DATABASE = "DEFAULT";

    private static final Map<String, ThreadLocal<Connection>> connections = new HashMap<String, ThreadLocal<Connection>>();

    /**
     * Sets the database connection for <var>database</var>.
     */
    public synchronized static void setConnection(String database,
            Connection newConnection) {
        ThreadLocal<Connection> conn = connections.get(database);
        if (conn == null) {
            conn = new ThreadLocal<Connection>();
            connections.put(database, conn);
        }
        conn.set(newConnection);
    }

    /**
     * Clears the database connection for <var>database</var>.
     */
    public static void clearConnection(String database) {
        setConnection(database, null);
    }

    /**
     * Returns the database connection for <var>database</var> or <code>null</code>, if no database
     * connection is available for this database.
     */
    public synchronized static Connection getConnection(String database) {
        final ThreadLocal<Connection> conn = connections.get(database);
        return conn != null ? conn.get() : null;
    }

    /**
     * Returns the database connection for <var>database</var>.
     * 
     * @throw {@link IllegalStateException} if there is no database connection for <var>database</var>
     */
    public synchronized static Connection checkAndGetConnection(String database) 
            throws IllegalStateException {
        final ThreadLocal<Connection> conn = connections.get(database);
        final Connection connection = conn != null ? conn.get() : null;
        if (connection == null) {
            throw new IllegalStateException("No database connection for database " + database);
        }
        return connection;
    }

}
