package net.lemnik.eodsql.impl;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.text.ParseException;

import net.lemnik.eodsql.ArrayDataSetCache;
import net.lemnik.eodsql.DataSetCache;
import net.lemnik.eodsql.QueryParams;
import net.lemnik.eodsql.Select;
import net.lemnik.eodsql.TypeMapper;
import net.lemnik.eodsql.spi.MethodImplementation;
import net.lemnik.eodsql.spi.util.DataObjectBinding;
import net.lemnik.eodsql.spi.util.MapDataObjectBinding;
import net.lemnik.eodsql.spi.util.NoDataObjectBinding;

/**
 * A {@link DynamicParameterFactory} for SELECT query method invocations.
 * 
 * @author Bernd Rinn
 */
class SelectDynamicParameterFactory implements DynamicParameterFactory {

    enum CallSchema {
        SIMPLE(false, -1, -1, 0), 
        BASERETURNTYPE(false, 0, -1, 1), 
        FETCHSIZE(false, -1, 0, 1), 
        BASERETURNTYPE_FETCHSIZE(false, 0, 1, 2), 
        QUERYPARAMS(true, -1, -1, 1);

        private final boolean useQueryParams;

        private final int baseReturnTypePos;

        private final int fetchSizePos;

        private final int sqlPos;
        
        CallSchema(boolean useQueryParams, int baseReturnTypePos, int fetchSizePos,
                int sqlPos) {
            this.useQueryParams = useQueryParams;
            this.baseReturnTypePos = baseReturnTypePos;
            this.fetchSizePos = fetchSizePos;
            this.sqlPos = sqlPos;
        }

        private QueryParams getQueryParams(final Object[] args) {
            return (QueryParams) args[0];
        }

        Class<?> getBaseReturnType(Object[] args) {
            if (useQueryParams) {
                return getQueryParams(args).getBaseReturnType();
            } else if (baseReturnTypePos >= 0)
            {
                return (Class) args[baseReturnTypePos];
            } else {
                return MapDataObjectBinding.getStringObjectMapObjectType();
            }
        }

        int getFetchSize(Object[] args) {
            if (useQueryParams) {
                return getQueryParams(args).getFetchSize();
            } else if (fetchSizePos >= 0) {
                return (Integer) args[fetchSizePos];
            } else {
                return 0;
            }
        }
        
        boolean isDisconnected(Object[] args) {
            if (useQueryParams) {
                return getQueryParams(args).isDisconnected();
            } else {
                return false;
            }
        }
        
        boolean isReadOnly(Object[] args) {
            if (useQueryParams) {
                return getQueryParams(args).isReadOnly();
            } else {
                return false;
            }
        }
        
        String getSql(Object[] args) {
            return (String) args[sqlPos];
        }
        
    }

    private final CallSchema callSchema;
    
    private final ReturnType returnType;

    SelectDynamicParameterFactory(final CallSchema callSchema, final ReturnType returnType) {
        this.callSchema = callSchema;
        this.returnType = returnType;
    }

    public Annotation createAnnotation(final Object[] args) {
        return new Select() {

            public String value() {
                return "";
            }

            public String sql() {
                return getSql(args);
            }

            public boolean disconnected() {
                return callSchema.isDisconnected(args);
            }

            public boolean rubberstamp() {
                return false;
            }

            public boolean readOnly() {
                return callSchema.isReadOnly(args);
            }

            public Class<? extends DataSetCache> cache() {
                return ArrayDataSetCache.class;
            }

            public Class<? extends Annotation> annotationType() {
                return Select.class;
            }

            public int fetchSize() {
                return callSchema.getFetchSize(args);
            }

            public int into() {
                return 0;
            }

            public Class<? extends DataObjectBinding> resultSetBinding() {
                return NoDataObjectBinding.class;
            }

            @SuppressWarnings("unchecked")
            public Class<? extends TypeMapper>[] parameterBindings() {
                return new Class[0];
            }

            public int timeout() {
                return 0;
            }

            public int timeoutParameterIndex() {
                return 0;
            }

            @Override
            public int hashCode() {
                int hash = "Select".hashCode();
                hash = 37 * hash + sql().hashCode();
                hash = 37 * hash + (disconnected() ? 1 : 0);
                hash = 37 * hash + (readOnly() ? 1 : 0);
                hash = 37 * hash + fetchSize();
                return hash;
            }

            @Override
            public boolean equals(Object obj) {
                if (obj instanceof Select == false) {
                    return false;
                }
                Select other = (Select) obj;
                return sql().equals(other.sql())
                        && disconnected() == other.disconnected()
                        && readOnly() == other.readOnly()
                        && fetchSize() == other.fetchSize();
            }
        };

    }

    public String getSql(Object[] args) {
        return callSchema.getSql(args);
    }

    public MethodImplementation<? extends Annotation> createMethodImplementation(
            Method method, Annotation annotation, Class[] queryParamTypes,
            Type returnType, Object[] args) throws ParseException {
        return new SelectMethodImplementation((Select) annotation,
                queryParamTypes, returnType);
    }

    public Class createBaseReturnType(final Object[] args) {
        return callSchema.getBaseReturnType(args); 
    }

    public ReturnType getReturnType() {
        return returnType;
    }

}