package net.lemnik.eodsql.impl;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.text.ParseException;
import java.util.Collection;

import net.lemnik.eodsql.GeneratedKeys;
import net.lemnik.eodsql.TypeMapper;
import net.lemnik.eodsql.Update;
import net.lemnik.eodsql.spi.MethodImplementation;

/**
 * A {@link DynamicParameterFactory} for UPDATE statement method invocations.
 * 
 * @author Bernd Rinn
 */
class UpdateDynamicParameterFactory implements DynamicParameterFactory {

    private final Class<?> baseReturnType;

    private final boolean singleRow;

    private final int sqlParamPos;

    private final int baseReturnTypeParamPos;

    private final boolean batchUpdate;

    private final GeneratedKeys keys;

    UpdateDynamicParameterFactory(final Class<?> baseReturnType,
            final int baseReturnTypeParamPos, final boolean singleRow,
            final int sqlParamPos, final boolean batchUpdate,
            final GeneratedKeys keys) {
        this.baseReturnType = baseReturnType;
        this.baseReturnTypeParamPos = baseReturnTypeParamPos;
        this.singleRow = singleRow;
        this.sqlParamPos = sqlParamPos;
        this.batchUpdate = batchUpdate;
        this.keys = keys;
    }

    public Annotation createAnnotation(final Object[] args) {
        return new Update() {
            public boolean batchUpdate() {
                return batchUpdate;
            }

            public GeneratedKeys keys() {
                return keys;
            }

            @SuppressWarnings("unchecked")
            public Class<? extends TypeMapper>[] parameterBindings() {
                return new Class[0];
            }

            public String sql() {
                return getSql(args);
            }

            public String value() {
                return "";
            }

            public Class<? extends Annotation> annotationType() {
                return Update.class;
            }

            public int timeout() {
                return 0;
            }

            public int timeoutParameterIndex() {
                return 0;
            }

            @Override
            public int hashCode() {
                int hash = "Update".hashCode();
                hash = 37 * hash + sql().hashCode();
                hash = 37 * hash + (batchUpdate() ? 1 : 0);
                hash = 37 * hash + keys.hashCode();
                return hash;
            }

            @Override
            public boolean equals(Object obj) {
                if (obj instanceof Update == false) {
                    return false;
                }
                Update other = (Update) obj;
                return sql().equals(other.sql())
                        && batchUpdate() == other.batchUpdate()
                        && keys().equals(other.keys());
            }
        };

    }

    public String getSql(Object[] args) {
        return (String) args[sqlParamPos];
    }

    public MethodImplementation<? extends Annotation> createMethodImplementation(
            Method method, Annotation annotation, Class[] queryParamTypes,
            Type returnType, Object[] args) throws ParseException {

        if (batchUpdate) {
            final Object[] queryParams = (Object[]) args[args.length - 1];
            return new BatchUpdateMethodImplementation((Update) annotation,
                    queryParamTypes, getQueryElementParamTypes(queryParamTypes,
                            queryParams), returnType, method);
        } else {
            return new UpdateMethodImplementation((Update) annotation,
                    queryParamTypes, returnType);
        }
    }

    private Class[] getQueryElementParamTypes(Class[] queryParamTypes,
            Object[] queryParams) {
        final int parameterCount = queryParamTypes.length;
        final Class<?>[] classes = new Class<?>[parameterCount];

        for (int i = 0; i < parameterCount; i++) {
            if (queryParamTypes[i].isArray()) {
                classes[i] = queryParamTypes[i].getComponentType();
            } else if (Collection.class.isAssignableFrom(queryParamTypes[i])) {
                final Collection c = (Collection) queryParams[i];
                if (c.isEmpty()) {
                    classes[i] = Object.class;
                } else {
                    classes[i] = c.iterator().next().getClass();
                }
            } else {
                classes[i] = queryParamTypes[i];
            }
        }
        return classes;
    }

    public Class createBaseReturnType(final Object[] args) {
        return (baseReturnTypeParamPos < 0) ? baseReturnType
                : (Class) args[baseReturnTypeParamPos];
    }

    public ReturnType getReturnType() {
        return singleRow ? ReturnType.SINGLEROW : ReturnType.DATASET;
    }

}